;
; defringeflat.pro - Removes fringes from an image
;
; Copyright (C) 2005 Patricio Rojo
;
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
;                   MA 02110-1301, USA.
;

;+
; NAME:
;	DEFRINGEFLAT
;
; PURPOSE:
;	This function removes the fringes from a flat field.
;
; CATEGORY:
;	Pato's Array Data Reduction Potpourri.
;
; CALLING SEQUENCE:
;
;	defringedflat = DEFRINGEFLAT(Flat, Refper)
;
;
; INPUTS:
;    Flat:       Two dimensional array containing the flat field to
;                defringe.
;    Refper:     Approximate fringe period (in pixels) at position
;                COLREF, ROWREF.
;
; KEYWORD PARAMETERS:
;  ARRAY KEYWORDS
;    BINW:       Number of rows included in median average to improve
;                signal-to-noise ratio. Default is 10.
;    COLREF:     Reference image column for which REFPER value is
;                given. Default is middle column.
;    ROWREF:     Reference row for which REFPER value is
;                given. Default is middle row.
;    TRIM:       [Deprecated]. Use ITRIM instead. Keyword is going to
;                be disabled in a future releas.
;    ITRIM:      Ignore datapoints outside these boundaries. Format is
;                [x0, y0, x1, y1].
;    GOODPIX:    Mask of good pixels. Must have same shape as Flat
;                array. 1 is good pixel, 0 is bad. 
;    GPIXRAD:    Radius of median filter to apply whenever a good
;                pixel mask is supplied. The median value will replace
;                the bad value. Default is 5.
;
;  TRACE FITTING KEYWORDS
;    NOCOI:      Ignore cone-of-influence (COI) boundary when fitting
;                the trace. If this is not set, then trace fitting
;                will stop at the column where the trace passes
;                through the COI boundary.
;    SILENT:     Set this for a silent run without any kind of
;                progress output. Error messages will show, though.
;    FITNAME:    What kind of fit to do. Currently 'GFC',
;                'GFCest', 'GFCnb', 'GVC', and 'true' are
;                supported. Default is 'GFC'. See below for an
;                explanation of each method under TRACE FITTING METHOD.
;    OSAMP:      Trace-fit oversampling factor. Default is 10.
;    BLINDCONTINUE: Continue the specified number of column blindly
;                if a peak was not successfully found. It uses the
;                last known trace center to find the next one. If
;                none of this extra columns gives a successfull fit
;                then it stop looking in that direction. Default is 25
;                pixels.
;    KEEPBACK:   If set, do not subtract the background from the trace,
;                i.e. do not consider the wavelet's background as part
;                of the flat but as part of the fringe pattern
;                instead. This option also put zeroes outside the
;                local minima to avoid contamination from wavelets
;                from different amplitudes. This switch has no
;                effect on the methods without background.
;    TRACELIM:   Specify this to have a limit on the trace
;                value. It should be a two-element array. The center
;                of the trace won't go below TRACELIM[0] or above
;                TRACELIM[1]. If the trace tries going beyond these
;                values, then a backsearch is started using the
;                reference period as a reference. Note that this
;                keyword has not been tested thoroughly, please report
;                any bug that you might find.
;    PERUPDATE:  Set this to update the value of REFPER with the
;                fitted period in the initial position. Useful when
;                processing several flats in a loop.
;    SHOWTRACE:  Show the fitted trace after having calculated it for
;                each row.
;
;  WAVELET PARAMETERS KEYWORDS
;    WAVEMOTHER: Name of wavelet to use. Currently only Morlet, Paul,
;                or DOG, with WAVEPARAM 6, 4, or 2, respectively, are
;                supported.
;    WAVEPARAM:  Parameter of the wavelet, currently only some values
;                are supported, see WAVEMOTHER.
;    NOPAD:      Set this flag not to zero-pad the data before doing
;                the wavelet transform.
;
;  FRINGE PARAMETERS KEYWORDS
;    SMOOTHFRINGE: Set this to patch and smooth (see below) the
;                parameters of the trace in the wavelet space before
;                reconstructing.
;    NSIGMA:     Rejection threshold for fitted fringe parameters. If 
;                any parameter is more than NSIGMA standard deviations
;                different from its median value then patch it. It has
;                an effect only when SMOOTHFRINGE is
;                specified. Default is 1.5.
;    RADMEDFILTER: Radius of the median filter to apply when
;                discriminating bad parameter values. It has an effect
;                only when SMOOTHFRINGE is specified. Default is 4.
;    RADSMOOTH:  Radius of mean filter to apply when smoothing in the
;                parameter space. It has an effect only when
;                SMOOTHFRINGE is specified. Default is 9.
;
;  OPTIONAL OUTPUT KEYWORD
;    OTRIM:      Output trim. Same format as ITRIM. This area is given
;                as a maximum possible area, it does not include the
;                limits set by Cone of Influence (COI).
;    FRINGE:     Returns the two dimensional fringe that was
;                constructed from the wavelet fit.
;    PRMFIT:     Returns the parametric fit used internally to
;                reconstruct the fit. The order of each parameter is
;                dependent on the value of FITNAME keyword, it has no
;                effect when 'true' is selected. This keyword is
;                mostly intended for debugging.
;
; OUTPUTS:
;       This function returns the corrected flat (input Flat minus the
;       fitted fringe).
;
; SIDE EFFECTS:
;    Value of input parameter Refper is overwritten when using
;    PERUPDATE option.
;
; RESTRICTIONS:
;     Do no operate heavy machinery while using this:D.
;
; PROCEDURE:
;       Area outside TRIM is cropped.
;       A 1D median average of BINW rows is computed for each of the
;     rows. Let's call each of the rows obtained in this manner
;     a meanrow.
;       For each of those meanrows, the slope is subtracted
;     through a linear fit. Then, a wavelet transform is performed
;     according to the WAVEMOTHER and the WAVEPARAM keywords.
;       Starting from Refper at position COLREF and ROWREF, the trace
;     of prominent periods in the absolute value of the wavelet space
;     is followed along the columns. The trace profile is fit by a
;     FITNAME function. Repeating this procedure for each row, we
;     obtain a fitted set of parameters (e.g., center, width,
;     amplitude of a Gaussian) for each pixel in the array.
;       If SMOOTHFRINGE is set, then each of the parameters is
;     patched (compared against the median value using NSIGMA and
;     RADMEDFILTER) and painted (smoothed according to RADSMOOTH).
;       Then, the wavelet transform is reconstructed from the
;     (smoothed) parameters. Finally, an inverse wavelet transform is
;     performed to obtain the fringe pattern, which is then removed
;     from the flat.
;
; TRACE FITTING METHODS:
;       All of the following procedures are applied to the data points
;      within the local minima. Those minimas are found at both sides
;      of the profile which contains the extrapolated reference period.
;        The different alternatives for the FITNAME keyword are:
;       * "GFC": (default). After oversampling, a center is found by
;                  the highest point in the local profile. Then a
;                  Gaussian function is fit to the trace's cross
;                  section but with the center previously fixed. Both
;                  centers estimates will be exactly the same only
;                  when the profile is symmetric. Four parameters
;                  (height, center, width, background) are obtained
;                  and can be smoothed. A negative background makes no
;                  sense. Hence, if such is the fit at a particular
;                  trace position, then the fit is retried on that
;                  pixel using 'GFCnb'. Using this method will not
;                  fit every wavelet point perfectly, but is more
;                  robust against removing non-fringe patterns from
;                  the image.
;       * "GFCnb": Another modification to 'GFC', where there is
;                  no background parameter (height, center, and width
;                  only). For real fringes, this and 'GFC' option
;                  give similar results but for sine-wave generated
;                  fringes, this option seems to work better.
;       * "GVC": Modification of 'GFC'. The center parameter is
;                  now fitted through a Gaussian and not fixed by the
;                  highest value. It fallbacks to 'GVCnb' when
;                  the background is negative.
;       * "GVCnb": Modification of 'GVC', where there is no
;                  background parameter.
;       * "GFCest": Yet another modification of 'GFC' fit, but the
;                  parameters are only estimated. 'Center' is the
;                  coordinate where the trace peaks, 'height' is the
;                  trace's value at center, 'width' is one-fifth
;                  (chosen arbitrarily as a good approximation) of the
;                  separation between the local lower minima, and
;                  'background' is zero. There is not really any
;                  advantage in choosing this alternative instead of
;                  the above other than for speed or as a failproof
;                  alternative.
;       * "true": Once the peak of trace's cross section has been
;                  found, all the points between the local lower
;                  minima at the sides of the trace are used to
;                  reconstruct the fringe's wavelet. The main
;                  disadvantage of this approach is that every single
;                  feature of the wavelet around the peak period is
;                  going to be removed from the original image. This
;                  might include real features from the image that
;                  were not intended to be cleaned. Smoothing of
;                  parameters is not available.
;
; DEPENDS ON:
;     maxtrace.pro    Finds a trace in a 2 dimensional array.
;     waverecon.pro   Produces an inverse wavelet transform.
;     wavelet.pro     Do a forward or inverse wavelet
;                     transform.
;                     Routine by Torrence and Compo. It is
;                     available from the DEFRINGEFLAT homepage
;                     http://www.das.uchile.cl/~pato/sw/
;                     The latest version should be available directly
;                     from the authors at
;                     http://paos.colorado.edu/research/wavelets/
;
;
; EXAMPLE:
;	Read example flat and defringe it. Start with a reference period
;     of 40 pixels at the center of the array, some trimming. Use a
;     Gaussian fit, interpolate period in the wavelet transform to 10
;     times more points for improved trace fitting. Show the fitted
;     trace after finding it in every row. Patch and smooth each of
;     the parameters of the Gaussian fit (center, width, height, and
;     background).
;
;         IDL> flat  = readfits('example/flat.fits', h)
;         IDL> func  = 'GFC'
;         IDL> clean = defringeflat(flat, 40, binw=40, /show, $
;                                   trim=[10,145,950,950], osamp=10, $
;                                   /smooth, fitname=func, fringe=fringe)
;         IDL> tvscl, flat
;         IDL> tvscl, fringe
;         IDL> tvscl, clean
;
; MODIFICATION HISTORY:
; 	Written by:	Pato Rojo, Cornell.  Aug-Oct 2005
;			pato@astro.cornell.edu
;-

function findandpatch_defringe, data, radius, nsigma, noval
compile_opt idl2, hidden
;look for outliers and patch

med = median(data, radius*2, /double)
n = size(data,/dim) - radius - 1
difprm = data - med
outliers = where(abs(difprm) gt $
                 stddev(median(difprm[radius:n[0],radius:n[1]], 2))*nsigma $
                 or data eq noval)

if outliers[0] eq -1 then return, data

ret = data
ret[outliers] = med[outliers]

return, ret
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

pro pnp_defringe, prmfit, radsmooth, radmedfilter, $
                  nsigma, silent, noval
compile_opt idl2, hidden
;Patch and paint(smooth) each of the fitted parameters

if ~ silent then print, ' Correcting parameters', format='(a,$)'
for e = 0, (size(prmfit, /dim))[2]-1 do begin 
    if ~ silent then print, string(e, format='(i2)') + ".", format='(a,$)'
    ;Patch. Find and fix outliers
    smoothed = findandpatch_defringe(reform(prmfit[*, *, e]), $
                                     radmedfilter, nsigma, noval)

    ;Paint. Smooth parameters
    if ~ silent then print,".", format='(a,$)'
    ind = where(smoothed eq noval)
    if (ind[0] ne -1) then smoothed[ind] = !values.f_nan
    smoothed = smooth(smoothed, 2*radsmooth+1, /nan, /edge)
    x = where(~ finite(smoothed))
    if x[0] ne -1 then smoothed[x] = noval
    if ~ silent then print, ".", format='(a,$)'

    prmfit[*, *, e] = smoothed
endfor
if ~ silent then print, " "

end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

function maketrace_defringe, prm, lm, idx, period, fitname, noval
compile_opt idl2, hidden
;Rebuild trace

nwa = dblarr(n_elements(idx), n_elements(period))
cnt = prm[*,1]
nx   = (size(prm, /dim))[0]
nprm = (size(prm, /dim))[1]

for xi = 0, nx-1 do if (cnt[xi] ne noval) then break

switch fitname of
    'GFCnb':
    'GVCnb': begin
        for i = xi, nx-1 do begin
            if (cnt[i] eq noval) then continue
            l1 = lm[0,i]
            l2 = lm[1,i]
            gauss = prm[i,0] * exp(-0.5*((period-prm[i,1])/prm[i,2])^2)
            nwa[i,l1:l2] = gauss[l1:l2]
        endfor
        break
    end

    'GFC':
    'GVC':
    'GFCest': begin
        for i = xi, nx-1 do begin
            if (cnt[i] eq noval) then continue
            l1 = lm[0,i]
            l2 = lm[1,i]
            gauss = prm[i,0] * exp(-0.5*((period-prm[i,1])/prm[i,2])^2) + $
              prm[i,3]
            nwa[i,l1:l2] = gauss[l1:l2]
        endfor
        break
    end
    else: message, "Method '" + fitname + $
      "' not enabled in maketrace_defringe"
endswitch

return, nwa

end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

pro buildfringe_defringe, silent, prmfit, locmin, nwp, noval, $ ;border
                          cdelta, psi0, period, scale, fitname, $ ;parameters
                          fringe ;returns
compile_opt idl2, hidden
;rebuild fringe from fitted wavelet

nx = (size(prmfit, /dim))[0]
ny = (size(prmfit, /dim))[1]
idx = indgen(nx)

if size(flat, /type) eq 5 then fringe = dblarr(nx, ny) $
else fringe = fltarr(nx, ny)

if keyword_set(locmin) then dozero = 1 else dozero = 0

;reconstruct
if ~ silent then print," Reconstructing..."
for j = 0, ny-1 do begin
    ;rebuild trace
    if dozero then lm = reform(locmin[*,*,j]) $
    else lm = [0, n_elements(period)-1]#replicate(1.0,nx)
    nwa = maketrace_defringe(reform(prmfit[*,j,*]), lm, $
                             idx, period, fitname, noval)

    cnw = complex(nwa*cos(nwp[*,*,j]), nwa*sin(nwp[*,*,j]))

    ;rebuild row
    fringe[*,j] = waverecon(cnw, 1, scale, cdelta, psi0)
endfor

end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

pro tracefit_defringe, flat, binw, osamp, $ ;basics
                       rrefper, colref, rowref, $ ;reference
                       blindcontinue, doshowtrace, keepback, fitname, $
                       tracelim, silent, nocoi, $ ;flags
                       waveparam, wavemother, pad, noval, $ ;wavelet parameters
                       prmfit, nwp, period, scale, locmin ;returns
compile_opt idl2, hidden
;find fringe from fitted wavelet of data

if ~ silent then print,' Fitting trace for each row...'
nx = (size(flat, /dim))[0]
ny = (size(flat, /dim))[1]
cbin = binw/2
idx = indgen(nx)

w = wavelet(idx, 1, coi=coi, scale=scale, period=period, $
            mother=wavemother, param=waveparam, pad=pad)
nwp = dblarr(nx, n_elements(scale), ny)

if (strpos(fitname,'GFC') ne -1 or strpos(fitname,'GVC') ne -1) then begin
    if ~ silent then print, fitname, format='(a,$)'
    prmfit = dblarr(nx, ny, 4)
endif else $
  message, "Fitting method '" + fitname + "' not enabled in tracefit"

if keepback then begin
    locmin = dblarr(2, nx, ny)
    print, 'kb', format='(a,$)'
endif
print, ':', format='(a,$)'

finrow = ny - binw - 1
steprow = 1
for order = 0, 1 do begin
    refper = rrefper

    for j = rowref - cbin, finrow, steprow do begin
        ;median average, subtract slope, and nullify borders
        p = median(reform(flat[*, j:j+binw]), $
                   dimension=2, /double)
        a = linfit(idx, p)
        p = p - a[0] - a[1]*idx

        ;produce wavelet
        w = wavelet(p, 1, coi=coi, scale=scale, period=period, $
                    mother=wavemother, param=waveparam, pad=pad)
        if ~ silent then begin
            if steprow lt 0 then print, '<', format='(a,$)' $
            else print,'>', format='(a,$)'
        endif
        if nocoi then coi2max = 0 else coi2max = coi

        ;find trace
        prm = maxtrace(abs(w), period, colref, refper,                     $
                       osamp=osamp, fit=fitname, /expy, tracelim=tracelim, $
                       noval=noval, coiborders=coi2max,                    $
                       blindcontinue=blindcontinue, localmin=localmin)
        prmfit[*, j+cbin, *] = prm
        if keepback then locmin[*, *, j+cbin] = localmin

        ;show if requested
        if doshowtrace then begin
            yrng = [100,10]
            !p.position = [.1,.83,.95,.95]
            blank = replicate(" ",30)
            plot, p, /ystyle, /xstyle, xtickname=blank,                  $
              title=string('Row: trimx + bin/2 + ', j, format='(a,i0)'), $
              yticks=3
            !p.position = [.1,.18,.95,.83]
            ttrace = where(prmfit[*,j+cbin,1] ne 0)
            rpidx = (where(period gt refper))[0]
            mval = max(abs(w[colref-5:colref+5, rpidx-5:rpidx+5]))
            contour, abs(w), idx, period, /noerase, xtickn=blank, $
              /xstyle, zrange=[1.3*mval,0], /fill, nlevels=25, $
              /ystyle, ytitle='Period', yrange=yrng, /ytype
            plots, idx, coi, noclip=0, thick=2
            plots, idx[ttrace], prmfit[ttrace,j+cbin,1], noclip=0
            plots, idx, replicate(40,1024), psym=0, noclip=0
            !p.position = [.1,.06,.95,.18]
            plot, idx[ttrace], prmfit[ttrace,j+cbin,0], /ystyle, /xstyle, $
              ytitle='height', xtitle='Pixel', /noerase, yticks=3
            !p.position = 0
        endif

        ;store phase
        nwp[*, *, j+cbin] = atan(w, /phase)
        refper = prmfit[colref, j+cbin, 1]
    endfor

    finrow  = 0
    steprow = -1
endfor

if ~ keepback then prmfit[*,*,3] = 0
if ~ silent then print,''

end


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


pro trueshape_defringe, flat, binw, osamp, rrefper, colref, rowref, $
                        waveparam, wavemother, pad, cdelta, psi0, fitname,  $
                        blindcontinue, doshowtrace, backsub, tracelim, $
                        silent, nocoi, midperiods, fringe
compile_opt idl2, hidden
;true shape search of fringe's wavelet

if ~ silent then print, ' Fitting trace for each row...'
nx = (size(flat, /dim))[0]
ny = (size(flat, /dim))[1]

midperiods = dblarr(ny)
if size(flat, /type) eq 5 then fringe = dblarr(nx, ny) $
else fringe = fltarr(nx,ny)

cbin = binw/2
idx = indgen(nx)

finrow = ny - binw - 1
steprow = 1
if ~ silent then print,'TS:',format='(a,$)'
for order = 0, 1 do begin
    refper = rrefper

    for j = rowref - cbin, finrow, steprow do begin

        ;take medianized row, subtract slope, nullify borders
        p = median(reform(flat[*, j:j+binw]), $
                   dimension=2, /double)
        a = linfit(idx, p)
        p = p - a[0] - a[1]*idx

        ;produce wavelet
        w = wavelet(p, 1, coi=coi, scale=scale, period=period, $
                    mother=wavemother, param=waveparam, pad=pad)

        if ~ silent then begin
            if steprow lt 0 then print, '<', format='(a,$)' $
            else print, '>', format='(a,$)'
        endif
        if nocoi then coi2max = 0 else coi2max = coi

        ;find trace
        nwa = maxtrace(abs(w), period, colref, refper,                    $
                       osamp=osamp, fitname=fitname, /expy, peaks=peaks,  $
                       trueback=trueback, tracelim=tracelim,              $
                       coiborders=coi2max, blindcontinue=blindcontinue,   $
                       localmin=locmin)

        ;show if requested
        if doshowtrace then begin
            !p.position = [.1,.05,.95,.8]
            rpidx = (where(period gt refper))[0]
            mval = max(nwa[colref-5:colref+5, rpidx-5:rpidx+5])
            contour, nwa, idx, period, $
              /xstyle, xtitle='Pixel', /ystyle, ytitle='Period',   $
              yrange=[100,10], /ytype, zrange=[1.3*mval,0], /fill, $
              nlevels=25
            plots, idx, coi, noclip=0, thick=2
            plots, idx, peaks, psym=0, noclip=0
            !p.position = [.1,.8,.95,.95]
            plot, p, /ystyle, /xstyle, xtickname=replicate(" ",30), $
              /noerase, title='Row '+string(j),                     $
              yrange=mean(p)+[-1,1]*4*stddev(p)
            !p.position = 0
        endif

        nwp = atan(w, /phase)

        if backsub then begin
            for x=0, nx-1 do begin
                lm = locmin[*,x]
                if nwa[x,total(lm)/2.0] eq 0.0 then continue
                nwa[x,lm[0]:lm[1]] = nwa[x,lm[0]:lm[1]] - trueback[x,0] - $
                  period[lm[0]:lm[1]]*trueback[x,1]
            endfor
        endif
        cnw = complex(nwa*cos(nwp), nwa*sin(nwp))

        ;rebuild row
        fringe[*, j+cbin] = waverecon(cnw, 1, scale, $
                                      cdelta, psi0)

        ;adjust refper for the next row and returning centers
        refper = peaks[colref]
        midperiods[j] = refper
    endfor

    finrow = 0
    steprow = -1
endfor
if ~ silent then print,''

end


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


function defringeflat, origflat, refper,                           $
                       trim=trim, itrim=itrim, otrim=otrim,        $
                       binwidth=binw, osamp=osamp,                 $
                       wavemother=wavemother, waveparam=waveparam, $
                       colref=colref, rowref=rowref,               $
                       keepback=keepback, silent=silent,           $
                       showtrace=showtrace, fitname=fitname,       $
                       smoothfringe=smoothfringe, nsigma=nsigma,   $
                       radmedfilter=radmedfilter,                  $
                       radsmooth=radsmooth, nopad=nopad,           $
                       goodpix=goodpix, gpixrad=gpixrad,           $
                       perupdate=perupdate, tracelim=tracelim,     $
                       fringe=fringe, nocoi=nocoi,                 $
                       blindcontinue=blindcontinue, prmfit=prmfit
compile_opt idl2
;Defringe a flat

on_error, 2
if ~ keyword_set(origflat) then message, 'No valid image was submitted'

nx = (size(origflat, /dim))[0]
ny = (size(origflat, /dim))[1]
if (size(origflat, /n_dimen) ne 2) then $
  message,"First parameter of defringeflat is not a two-dimensional array"

if keyword_set(trim) and ~ keyword_set(itrim) then begin
    itrim = trim
    print, 'WARNING: defringeflat.pro:: you are still using deprecated' + $
      ' keyword TRIM, use ITRIM instead'
endif
if ~ keyword_set(itrim) then        otrim = [0, 0, nx-1, ny-1] $
else if n_elements(itrim) eq 4 then $
  otrim = (itrim < [nx-1, ny-1, nx-1, ny-1]) > [0, 0, 0, 0] $
else message, "Trim parameter is invalid. It need to be a 4 elements array"
on_error, 0

pad     = 1
if keyword_set(nopad)    then pad     = 0

if ~ keyword_set(keepback)     then keepback      = 0
if ~ keyword_set(nocoi)        then nocoi         = 0
if ~ keyword_set(silent)       then silent        = 0
if ~ keyword_set(nsigma)       then nsigma        = 1.5
if ~ keyword_set(radmedfilter) then radmedfilter  = 4
if ~ keyword_set(radsmooth)    then radsmooth     = 9
if ~ keyword_set(showtrace)    then showtrace     = 0
if ~ keyword_set(wavemother)   then wavemother    = 'Morlet'
if ~ keyword_set(waveparam)    then waveparam     = 6
if ~ keyword_set(fitname)      then fitname       = 'GFC'
if ~ n_elements(binw)          then binw          = 11
if ~ n_elements(osamp)         then osamp         = 10
if ~ n_elements(colref)        then colref        = nx/2
if ~ n_elements(rowref)        then rowref        = ny/2
if ~ n_elements(blindcontinue) then blindcontinue = 25

switch wavemother of
    'Morlet': begin
        if waveparam eq 6 then begin
            Cdelta = 0.776
            psi0 = !PI^(-0.25)
            break
        endif
    end
    'Paul': begin
        if waveparam eq 4 then begin
            Cdelta = 1.132
            psi0 = 2.^waveparam * FACTORIAL(waveparam) / $
              SQRT(!PI * FACTORIAL(2*waveparam))
            break
        endif
    end
    'Dog': begin
        if waveparam eq 2 then begin
            Cdelta = 3.541
            psi0 = 0.867325
            break
        endif
    end
    else: $
      message, string("Cannot do a wavelet of type " + wavemother + $
                      " (parameter ", waveparam, "), try Morlet, Paul, or" + $
                      " DOG only, with parameters 6, 4, or 2, repectively", $
                      format='(a,i0,a)')
endswitch

;if a good pixel mask is supplied, replace those values for their
;median value.
flat = origflat
if keyword_set(goodpix) then begin
    if ~ keyword_set(gpixrad) then gpixrad = 5
    idx = where(goodpix eq 0)
    flat[idx] = (median(flat, gpixrad))[idx]
endif
flat = flat[otrim[0]:otrim[2], otrim[1]:otrim[3]]
colr = colref - otrim[0]
rowr = rowref - otrim[1]

if strpos(fitname, 'true') ne -1 then begin
    trueshape_defringe, flat, binw, osamp, refper, colr, rowr, $
      waveparam, wavemother, pad, cdelta, psi0, fitname, blindcontinue, $
      showtrace, ~ keepback, tracelim, silent, nocoi, midperiods, $
      frin
    if keyword_set(perupdate) then rrefper = midperiods[rowr]
    fringe = origflat*0.0
    fringe[otrim[0]:otrim[2], otrim[1]:otrim[3]] = frin
    return, origflat - fringe
endif

;lookfor trace to fit
tracefit_defringe, flat, binw, osamp, refper, colr, rowr, blindcontinue, $
  showtrace, keepback, fitname, tracelim, silent, nocoi, waveparam,       $
  wavemother, pad, noval, prmfit, nwp, period, scale, locmin
if keyword_set(perupdate) then refper = prmfit[colr, rowr, 1]
n = [binw/2, (size(flat,/dim))[1]-binw/2-1]
flat   = flat  [*, n[0]:n[1]]
prmfit = prmfit[*, n[0]:n[1], *]
if keyword_set(locmin) then locmin = locmin[*, *, n[0]:n[1]]
nwp    = nwp   [*, *, n[0]:n[1]]
otrim += [0, 1, 0, -1] * binw/2

if keyword_set(smoothfringe) then begin
                                ;patch, paint
    pnp_defringe, prmfit, radsmooth, radmedfilter, nsigma, silent, noval
    n = (size(flat, /dim)) - radsmooth - 1
    flat   = flat  [radsmooth:n[0], radsmooth:n[1]]
    nwp    = nwp   [radsmooth:n[0], *, radsmooth:n[1]]
    prmfit = prmfit[radsmooth:n[0], radsmooth:n[1], *]
    if keyword_set(locmin) then $
      locmin = locmin[*, radsmooth:n[0], radsmooth:n[1]]
    otrim += [1, 1, -1, -1] * radsmooth
endif

;build fringe pattern
buildfringe_defringe, silent, prmfit, locmin, nwp, noval, cdelta, psi0, $
  period, scale, fitname, frin

fringe = origflat*0.0
fringe[otrim[0]:otrim[2], otrim[1]:otrim[3]] = frin
return, origflat - fringe

end
